/* Styling of the dropdowns */
select {
    appearance: none;
    border: 0;
    flex: 1;
    padding: 0 1em;
    background-color: #eee;
    cursor: pointer;
}
.select {
    box-shadow: 0 5px 5px rgba(0, 0, 0, 0.25);
    cursor: pointer;
    display: flex;
    width: 100px;
    height: 25px;
    border-radius: .25em;
    overflow: hidden;
    position: relative;
    margin-right: 15px;
}
.select::after {
    position: absolute;
    content: '\25BC';
    font-size: 9px;
    top: 0;
    right: 0;
    padding: 1em;
    background-color: #ddd;
    transition: .25s all ease;
    pointer-events: none;
}
.select:hover::after {
    color: black;
}
#dropdown-container {
    margin-left: 10px;
    margin-top: 20px;
    display: flex;
    flex-wrap: wrap;
}